/*
 *	The Killer Daemon
 *
 *	(c) 1997--1998, 2006 Martin Mares <mj@ucw.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU General Public License. See file COPYING in any of the GNU packages.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <dirent.h>
#include <stdarg.h>
#include <pwd.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <utmp.h>
#include <signal.h>
#include <errno.h>

#define CONFIG_FILE "/etc/killerd.conf"

typedef unsigned char byte;
typedef unsigned int ulg;
typedef unsigned short word;

struct kill_node {
  struct kill_node *next;
  ulg data;
  byte name[1];
};

#define NC_REPEAT 1			/* Repeat period setting */
#define NC_EXCLUDE_USER 2		/* This user should be excluded */
#define NC_INTERACTIVE 3		/* This program is interactive */
#define NC_IDLE_LIMIT 4			/* Maximal allowed idle time */
#define NC_IDLE_WARN 5			/* Time between idle warn message and kill */
#define NC_SESSION_LIMIT 6		/* Maximal time of login session */
#define NC_SESSION_WARN 7		/* Time between session length warn message and kill */
#define NC_TIME_WARN_MSG 8		/* Warning message because of session time limit */
#define NC_IDLE_WARN_MSG 9		/* Warning message because of idle time limit */
#define NC_TASK_RUN_LIMIT 10		/* CPU time limit for tasks */
#define NC_TASK_TIME_LIMIT 11		/* Wall clock time limit for tasks */
#define NC_LOG_FILE_NAME 12		/* Name of log file */
#define NC_EXCLUDE_PROCESS 13		/* Process to be excluded */
#define NC_ONLY_LOGIN 14		/* This program should be killed unless the user is logged in */
#define NC_MAX 15

static char nc_names[] = "#RXUIiSsmMtTLPD";

static struct kill_node *first_node[NC_MAX];

static int debugging;
static time_t now;

static void __attribute__ ((noreturn))
err(char *c, ...)
{
  va_list args;

  va_start(args, c);
  fputs("killerd: ", stderr);
  vfprintf(stderr, c, args);
  fputs("!\n", stderr);
  exit(1);
}

static char *
when(void)
{
  struct tm *tm;
  static char buf[32];

  now = time(NULL);
  tm = localtime(&now);
  sprintf(buf, "%02d-%02d-%04d %02d:%02d:%02d",
	  tm->tm_mday, tm->tm_mon+1, tm->tm_year + 1900,
	  tm->tm_hour, tm->tm_min, tm->tm_sec);
  return buf;
}

static void *
xmalloc(ulg size)
{
  void *k = malloc(size);
  if (!k)
    err("Out of memory");
  return k;
}

/* Getting at the value of HZ */

#include <asm/param.h>

static ulg kernel_HZ;

#ifndef AT_CLKTCK
#define AT_CLKTCK 17
#endif

static void
init_HZ(void)
{
  char *how = NULL;

  /* Find the ELF auxiliary vector and search for the AT_CLKTCK note. */
  unsigned long *env = (unsigned long *) __environ;
  while (*env++)
    ;
  while (*env && env[0] != AT_CLKTCK)
    env += 2;
  if (*env)
    {
      kernel_HZ = env[1];
      how = "AT_CLKTCK";
    }

  if (!kernel_HZ)
    {
#ifdef USER_HZ
      kernel_HZ = USER_HZ;
      how = "USER_HZ";
#elif defined(HZ)
      kernel_HZ = HZ;
      how = "HZ";
#else
      err("Cannot determine the kernel clock tick frequency");
#endif
    }

  if (debugging)
    fprintf(stderr, "Guessed HZ=%d (%s)\n", kernel_HZ, how);
}

static void
add_kill_node(char *name, ulg cat, ulg data)
{
  struct kill_node *n;

  n = xmalloc(sizeof(struct kill_node) + strlen(name));
  n->next = first_node[cat];
  first_node[cat] = n;
  n->data = data;
  strcpy(n->name, name);
}

static struct kill_node *
find_kill_node2(struct kill_node *n, char *name)
{
  while (n)
    {
      if (!strcmp(n->name, name))
	return n;
      n = n->next;
    }
  return NULL;
}

static struct kill_node *
find_kill_node(char *name, ulg cat)
{
  struct kill_node *n;

  if (n = find_kill_node2(first_node[cat], name ? : ""))
    return n;
  return find_kill_node2(first_node[cat], "*");
}

static void
load_conf(char *name)
{
  FILE *f;
  char line[256];
  char *c, *d;
  ulg val;

  if (!(f = fopen(name, "r")))
    err("Cannot open %s: %m", name);
  for(;;)
    {
      fgets(line, sizeof(line), f);
      if (feof(f))
	break;
      c = strchr(line, '\n');
      if (!c)
	err("Line too long");
      *c = 0;
      if (!line[0] || line[0] == '#')
	continue;
      c = line+1;
      while (*c && *c != '=')
	c++;
      if (!*c)
	val = 0;
      else
	{
	  *c++ = 0;
	  val = atol(c);
	}
      if (! (d = strchr(nc_names, line[0])))
	err("Unknown option: %c", line[0]);
      add_kill_node(line+1, d - nc_names, val);
    }
  fclose(f);
}

struct user_node {
  struct user_node *next;
  uid_t uid;
  pid_t pid;
  int idle, logged_in, excluded;
  char username[UT_NAMESIZE+2];
  char linename[UT_LINESIZE+2];
};

struct user_node *free_user_nodes;
struct user_node *first_user_node;

static void
strcpy_padded(char *to, char *from, int size)
{
  memcpy(to, from, size);
  to[size] = 0;
}

static void
scan_users(void)
{
  struct utmp *ut;
  struct passwd *pw;
  struct user_node *u;
  struct stat st;
  char device[32];

  utmpname(UTMP_FILE);
  while (ut = getutent())
    if (ut->ut_type == USER_PROCESS && ut->ut_user[0])
      {
	if (free_user_nodes)
	  {
	    u = free_user_nodes;
	    free_user_nodes = u->next;
	  }
	else
	  u = xmalloc(sizeof(struct user_node));
	u->next = first_user_node;
	first_user_node = u;
	u->logged_in = now - ut->ut_time;
	if (u->logged_in < 0)
	  u->logged_in = 0;
	strcpy_padded(u->username, ut->ut_user, UT_NAMESIZE);
	pw = getpwnam(u->username);
	if (pw)
	  u->uid = pw->pw_uid;
	else
	  u->uid = -1;
	u->pid = ut->ut_pid;
	strcpy_padded(u->linename, ut->ut_line, UT_LINESIZE);
	strcpy(device, "/dev/");
	strcat(device, u->linename);
	if (stat(device, &st))
	  u->idle = -1;
	else
	  {
	    time_t la = st.st_atime;
#if 0
	    if (la < st.st_mtime)
	      la = st.st_mtime;
	    if (la < st.st_ctime)
	      la = st.st_ctime;
#endif
	    u->idle = (la > now) ? 0 : now - la;
	  }
	u->excluded = !!find_kill_node(u->username, NC_EXCLUDE_USER);
	if (debugging)
	  fprintf(stderr, "/%s/ uid=%d logtime=%d idle=%d dev=%s ex=%d\n", u->username, u->uid, u->logged_in,
		  u->idle, u->linename, u->excluded);
      }
  endutent();
}

static void
free_users(void)
{
  struct user_node *k;

  while (first_user_node)
    {
      k = first_user_node;
      first_user_node = first_user_node -> next;
      k->next = free_user_nodes;
      free_user_nodes = k;
    }
}

struct process_descr {
  int pid;
  int session;				/* STAT */
  int ruid;				/* OWNER */
  int tty;				/* STAT */
  char name[256];			/* STAT */
  int utime, stime, cutime, cstime, start_jiffies, nice; /* STAT */
};

#define GPI_STAT 1
#define GPI_OWNER 2

static int
get_process_info(pid_t pid, struct process_descr *p, int flags)
{
  int fil, z, trash;
  char buf[1024];
  char *k, *m;
  int oook = 1;
  FILE *f;

  p->pid = pid;

  if (flags & GPI_STAT)
    {
      sprintf(buf, "/proc/%d/stat", pid);
      fil = open(buf, O_RDONLY);
      if (fil >= 0)
	{
	  z = read(fil, buf, 1022);
	  if (z > 0 && z < 1022)
	    {
	      buf[z] = 0;
	      m = strchr(buf, '(');
	      if (m)
		k = strchr(m+1, ')');
	      if (m && k)
		{
		  m++;
		  *k++ = 0;
		  if (strlen(m) < 255)
		    strcpy(p->name, m);
		  else
		    strcpy(p->name, "<name-too-long>");
		  if (sscanf(k+3, "%d%d%d%d%d%d%d%d%d%d%d%d%d%d%d%d%d%d%d",
			     &trash, &trash, &p->session, &p->tty, &trash, &trash, &trash, &trash,
			     &trash, &trash, &p->utime, &p->stime, &p->cutime, &p->cstime, &trash,
			     &p->nice, &trash, &trash, &p->start_jiffies) != 19)
		    {
		      fprintf(stderr, "Error parsing /proc/%d/stat\n", pid);
		      oook = 0;
		    }
		}
	      else
		{
		  fprintf(stderr, "Syntax error in /proc/%d/stat\n", pid);
		  oook = 0;
		}
	    }
	  else
	      /* Error not reported as the process might have disappeared in the meantime */
	      oook = 0;
	  close(fil);
	}
      else
	oook = 0;
    }

  if (flags & GPI_OWNER)
    {
      sprintf(buf, "/proc/%d/status", pid);
      f = fopen(buf, "r");
      p->ruid = -1;
      if (f)
	{
	  for(;;)
	    {
	      fgets(buf, 254, f);
	      if (feof(f))
		break;
	      k = strchr(buf, '\n');
	      if (k)
		*k = 0;
	      if (!strncmp(buf, "Uid:", 4))
		sscanf(buf+5, "%d", &p->ruid);
	    }
	  fclose(f);
	}
      else
	oook = 0;
    }

  return oook;
}

static unsigned long reference_jiffies;
static time_t reference_time;

static void
init_jiffie_clock(void)
{
  struct process_descr us;

  if (!get_process_info(getpid(), &us, GPI_STAT))
    err("Cannot identify myself");
  reference_jiffies = us.start_jiffies;
  reference_time = time(NULL);
}

static void
kill_him(int pid)
{
  if (debugging)
    {
      fprintf(stderr, "Would send kill to %d\n", pid);
      return;
    }
  if (kill(pid, SIGHUP))
    return;
  sleep(1);
  if (kill(pid, 0))
    return;
  sleep(5);
  if (kill(pid, SIGTERM))
    return;
  sleep(1);
  if (kill(pid, 0))
    return;
  sleep(5);
  kill(pid, SIGKILL);
}

static void
kick_user(struct user_node *u, char *why)
{
  struct process_descr pd;

  fprintf(stderr, "%s: Kicking out user %s at %s (pid %d) [%s]\n",
	  when(), u->username, u->linename, u->pid, why);
  if (!get_process_info(u->pid, &pd, GPI_STAT))
    return;
  kill_him(-pd.session);
}

static void
warn_user(struct user_node *u, ulg msgn)
{
  struct kill_node *n = first_node[msgn];
  char *msg = "\n\n\a\a*** Warning: You will be logged out automatically soon ***\n\n";
  char devname[32], msgbuf[256], *c;
  int k, r;

  if (n)
    msg = n->name;
  sprintf(devname, "/dev/%s", u->linename);

  if (debugging)
    {
      fprintf(stderr, "%s at %s is being warned\n", u->username,
	      devname);
      return;
    }

  c = msgbuf;
  while (*msg)
    {
      if (*msg == '\\')
	{
	  switch (*++msg)
	    {
	    case 'a':
	      *c++ = '\a';
	      break;
	    case 'n':
	      *c++ = '\n';
	      break;
	    case 'r':
	      *c++ = '\r';
	      break;
	    default:
	      *c++ = *msg;
	    }
	}
      else
	*c++ = *msg;
      msg++;
    }
  *c = 0;

  k = open(devname, O_WRONLY | O_NOCTTY | O_NDELAY);
  if (k > 0)
    {
      msg = msgbuf;
      while (*msg)
	{
	  r = write(k, msg, strlen(msg));
	  if (r > 0)
	    msg += r;
	  else if (errno == EWOULDBLOCK)
	    sleep(10);
	  else
	    break;
	}
      close(k);
    }
}

static void
do_users(void)
{
  struct user_node *u;
  struct kill_node *n;
  ulg session_limit = -1000000;
  ulg session_warn = 120;
  ulg idle_limit = -1000000;
  ulg idle_warn = 120;
  ulg user_kick_time, user_warn_time;

  if (n = find_kill_node(NULL, NC_SESSION_LIMIT))
    session_limit = n->data;
  if (n = find_kill_node(NULL, NC_SESSION_WARN))
    session_warn = n->data;
  if (n = find_kill_node(NULL, NC_IDLE_LIMIT))
    idle_limit = n->data;
  if (n = find_kill_node(NULL, NC_IDLE_WARN))
    idle_warn = n->data;

  for(u=first_user_node; u; u=u->next)
    {
      if (u->excluded)
	continue;
      user_warn_time = (n = find_kill_node(u->username, NC_SESSION_LIMIT)) ? n->data : session_limit;
      user_kick_time = user_warn_time + ((n = find_kill_node(u->username, NC_SESSION_WARN)) ? n->data : session_warn);
      if (u->logged_in > user_kick_time)
	kick_user(u, "session time exceeded");
      else if (u->logged_in > user_warn_time)
	warn_user(u, NC_TIME_WARN_MSG);
      user_warn_time = (n = find_kill_node(u->username, NC_IDLE_LIMIT)) ? n->data : idle_limit;
      user_kick_time = user_warn_time + ((n = find_kill_node(u->username, NC_IDLE_WARN)) ? n->data : idle_warn);
      if (u->idle > user_kick_time)
	kick_user(u, "idle time exceeded");
      else if (u->idle > user_warn_time)
	warn_user(u, NC_IDLE_WARN_MSG);
    }
}

static void
kill_process(struct process_descr *pd, char *why, char *name)
{
  fprintf(stderr, "%s: Killing process %s (pid %d, user %s) [%s]\n",
	  when(), pd->name, pd->pid, name, why);
  kill_him(pd->pid);
}

static int
user_logged_in(uid_t uid)
{
  struct user_node *u;
  
  for(u=first_user_node; u; u=u->next)
    if (uid != u->uid)
      return 1;
  return 0;
}

static void
do_procs(void)
{
  DIR *d;
  struct dirent *e;
  struct kill_node *n;
  struct passwd *p;
  ulg max_run_time = ~0;
  ulg max_task_time = ~0;
  ulg mrt, mtt, rut;

  if (n = find_kill_node(NULL, NC_TASK_RUN_LIMIT))
    max_run_time = n->data;
  if (n = find_kill_node(NULL, NC_TASK_TIME_LIMIT))
    max_task_time = n->data;

  if (d = opendir("/proc"))
    {
      while (e = readdir(d))
	{
	  int pid;
	  struct process_descr pd;

	  sscanf(e->d_name, "%d", &pid);
	  if (get_process_info(pid, &pd, GPI_STAT | GPI_OWNER))
	    {
	      if (debugging)
		fprintf(stderr, "%d /%s/ %d %d %d %d %d %d %d %d %d\n", pid, pd.name, pd.session, pd.tty, pd.ruid,
			pd.utime, pd.stime, pd.cutime, pd.cstime, pd.start_jiffies, pd.nice);
	      if (find_kill_node(pd.name, NC_EXCLUDE_PROCESS))
		continue;
	      p = getpwuid(pd.ruid);
	      if (!p)
		{
		  fprintf(stderr, "UID %d not in passwd!\n", pd.ruid);
		  continue;
		}
	      if (find_kill_node(p->pw_name, NC_EXCLUDE_USER))
		continue;
	      if (!pd.tty && find_kill_node(pd.name, NC_INTERACTIVE))
		{
		  kill_process(&pd, "interactive", p->pw_name);
		  continue;
		}
	      if (find_kill_node(pd.name, NC_ONLY_LOGIN) && !user_logged_in(pd.ruid))
		{
		  kill_process(&pd, "owner not logged in", p->pw_name);
		  continue;
		}
	      mrt = (n = find_kill_node(p->pw_name, NC_TASK_RUN_LIMIT)) ? n->data : max_run_time;
	      mtt = (n = find_kill_node(p->pw_name, NC_TASK_TIME_LIMIT)) ? n->data : max_task_time;
	      if ((pd.utime + pd.stime) / kernel_HZ >= mrt)
		{
		  kill_process(&pd, "CPU time exceeded", p->pw_name);
		  continue;
		}
	      if (pd.start_jiffies < reference_jiffies)
		{			/* The trick is here to have true 32-bit jiffies */
		  rut = reference_jiffies - pd.start_jiffies;
		  rut = reference_time - rut / kernel_HZ;
		}
	      else
		{
		  rut = pd.start_jiffies - reference_jiffies;
		  rut = reference_time + rut / kernel_HZ;
		}
	      if (now - rut >= mtt && now > rut)
		{
		  kill_process(&pd, "run time exceeded", p->pw_name);
		  continue;
		}
	    }
	}
      closedir(d);
    }
}

int
main(int argc, char **argv)
{
  struct kill_node *n;
  int repeat_time = 60;
  int pid;

  if (argc > 1)
    {
      debugging = 1;
      load_conf(argv[1]);
    }
  else
    load_conf(CONFIG_FILE);

  if (n = first_node[NC_REPEAT])
    repeat_time = n->data;

  if (n = first_node[NC_LOG_FILE_NAME])
    {
      int f = open(n->name, O_RDWR | O_APPEND | O_CREAT, 0666);
      if (f < 0)
	{
	  perror("Open log file");
	  return 1;
	}
      close(2);
      dup(f);
      close(f);
    }

  if (!debugging)
    {
      pid = fork();
      if (pid < 0)
        {
          perror("fork");
	  return 1;
        }
      if (pid)
        return 0;
    }

  fprintf(stderr, "%s: Started\n", when());
  init_HZ();
  init_jiffie_clock();
  for(;;)
    {
      now = time(NULL);
      scan_users();
      do_users();
      do_procs();
      free_users();
      if (debugging)
	return 1;
      fflush(stderr);
      sleep(repeat_time);
    }
}
